# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Iterable, List

from troubadix.plugin import Plugin

from .badwords import CheckBadwords
from .copyright_text import CheckCopyrightText
from .copyright_year import CheckCopyrightYear
from .creation_date import CheckCreationDate
from .cve_format import CheckCVEFormat
from .cvss_format import CheckCVSSFormat
from .dependencies import CheckDependencies
from .dependency_category_order import CheckDependencyCategoryOrder
from .deprecated_dependency import CheckDeprecatedDependency
from .deprecated_functions import CheckDeprecatedFunctions
from .description import CheckDescription
from .double_end_points import CheckDoubleEndPoints
from .duplicated_script_tags import CheckDuplicatedScriptTags
from .duplicate_oid import CheckDuplicateOID
from .encoding import CheckEncoding
from .forking_nasl_functions import CheckForkingNaslFunctions
from .get_kb_on_services import CheckGetKBOnServices
from .grammar import CheckGrammar
from .http_links_in_tags import CheckHttpLinksInTags
from .illegal_characters import CheckIllegalCharacters
from .log_messages import CheckLogMessages
from .misplaced_compare_in_if import CheckMisplacedCompareInIf
from .missing_desc_exit import CheckMissingDescExit
from .missing_tag_solution import CheckMissingTagSolution
from .newlines import CheckNewlines
from .no_solution import CheckNoSolution
from .overlong_script_tags import CheckOverlongScriptTags
from .prod_svc_detect_in_vulnvt import CheckProdSvcDetectInVulnvt
from .qod import CheckQod
from .reporting_consistency import CheckReportingConsistency
from .risk_factor import CheckRiskFactor
from .script_add_preference_type import CheckScriptAddPreferenceType
from .script_calls_empty_values import CheckScriptCallsEmptyValues
from .script_calls_mandatory import CheckScriptCallsMandatory
from .script_calls_recommended import CheckScriptCallsRecommended
from .script_category import CheckScriptCategory
from .script_copyright import CheckScriptCopyright
from .script_family import CheckScriptFamily
from .script_tag_form import CheckScriptTagForm
from .script_tag_whitespaces import CheckScriptTagWhitespaces
from .script_version_and_last_modification_tags import (
    CheckScriptVersionAndLastModificationTags,
)
from .script_xref_form import CheckScriptXrefForm
from .script_xref_url import CheckScriptXrefUrl
from .security_messages import CheckSecurityMessages
from .set_get_kb_calls import CheckWrongSetGetKBCalls
from .solution_text import CheckSolutionText
from .solution_type import CheckSolutionType
from .spelling import CheckSpelling
from .tabs import CheckTabs
from .todo_tbd import CheckTodoTbd
from .trailing_spaces_tabs import CheckTrailingSpacesTabs
from .update_modification_date import UpdateModificationDate
from .using_display import CheckUsingDisplay
from .valid_oid import CheckValidOID
from .valid_script_tag_names import CheckValidScriptTagNames
from .variable_assigned_in_if import CheckVariableAssignedInIf
from .vt_placement import CheckVTPlacement


_PLUGINS = [
    CheckBadwords,
    CheckCopyrightText,
    CheckCopyrightYear,
    CheckCreationDate,
    CheckCVEFormat,
    CheckCVSSFormat,
    CheckDependencies,
    CheckDependencyCategoryOrder,
    CheckDeprecatedDependency,
    CheckDeprecatedFunctions,
    CheckDescription,
    CheckDoubleEndPoints,
    CheckDuplicatedScriptTags,
    CheckEncoding,
    CheckForkingNaslFunctions,
    CheckGetKBOnServices,
    CheckGrammar,
    CheckHttpLinksInTags,
    CheckIllegalCharacters,
    CheckLogMessages,
    CheckMisplacedCompareInIf,
    CheckMissingDescExit,
    CheckMissingTagSolution,
    CheckNewlines,
    CheckOverlongScriptTags,
    CheckProdSvcDetectInVulnvt,
    CheckQod,
    CheckReportingConsistency,
    CheckRiskFactor,
    CheckScriptAddPreferenceType,
    CheckScriptCallsEmptyValues,
    CheckScriptCallsMandatory,
    CheckScriptCallsRecommended,
    CheckScriptCategory,
    CheckScriptCopyright,
    CheckScriptFamily,
    CheckScriptTagForm,
    CheckScriptTagWhitespaces,
    CheckScriptVersionAndLastModificationTags,
    CheckScriptXrefForm,
    CheckScriptXrefUrl,
    CheckSecurityMessages,
    CheckSolutionText,
    CheckSolutionType,
    CheckSpelling,
    CheckTabs,
    CheckTodoTbd,
    CheckTrailingSpacesTabs,
    CheckUsingDisplay,
    CheckValidOID,
    CheckValidScriptTagNames,
    CheckVariableAssignedInIf,
    CheckVTPlacement,
    CheckWrongSetGetKBCalls,
]

_PRE_RUN_PLUGINS = [
    CheckDuplicateOID,
    CheckNoSolution,
]


class Plugins:
    def __init__(self, plugins: List[Plugin], prerun_plugins: List[Plugin]):
        self._plugins = plugins
        self._prerun_plugins = prerun_plugins

    def __len__(self) -> int:
        return len(self._plugins)

    def __iter__(self) -> Iterable[Plugin]:
        return iter(self._plugins)

    def get_prerun_plugins(self) -> Iterable[Plugin]:
        return self._prerun_plugins


class UpdatePlugins(Plugins):
    def __init__(self):
        super().__init__([UpdateModificationDate], [])


class StandardPlugins(Plugins):
    def __init__(
        self,
        excluded_plugins: List[str] = None,
        included_plugins: List[str] = None,
    ) -> None:
        super().__init__(_PLUGINS, _PRE_RUN_PLUGINS)

        if excluded_plugins:
            self._plugins = [
                plugin
                for plugin in self._plugins
                if plugin.__name__ not in excluded_plugins
                and plugin.name not in excluded_plugins
            ]
            self._prerun_plugins = [
                plugin
                for plugin in self._prerun_plugins
                if plugin.__name__ not in excluded_plugins
                and plugin.name not in excluded_plugins
            ]
        if included_plugins:
            self._plugins = [
                plugin
                for plugin in self._plugins
                if plugin.__name__ in included_plugins
                or plugin.name in included_plugins
            ]
            self._prerun_plugins = [
                plugin
                for plugin in self._prerun_plugins
                if plugin.__name__ in included_plugins
                or plugin.name in included_plugins
            ]



#  Copyright (c) 2022 Greenbone Networks GmbH
#
#  SPDX-License-Identifier: GPL-3.0-or-later
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from pathlib import Path
from typing import Iterator

from troubadix.helper.patterns import (
    ScriptTag,
    SpecialScriptTag,
    get_script_tag_pattern,
    get_special_script_tag_pattern,
)
from troubadix.plugin import FileContentPlugin, LinterError, LinterResult


class CheckScriptVersionAndLastModificationTags(FileContentPlugin):
    name = "check_script_version_and_last_modification_tags"

    def check_content(
        self,
        nasl_file: Path,
        file_content: str,
    ) -> Iterator[LinterResult]:
        """The script checks if the passed VT has a correct syntax of the
        following two tags:

        - script_version();
        - script_tag(name:"last_modification", value:"");

        An error will be thrown if the syntax of those two tags does not match
        the requirements.

        Args:
            nasl_file: The VT that shall be checked
            file_content: The content of the VT that shall be checked
        """
        if nasl_file.suffix == ".inc":
            return

        # script_version("2019-03-21T12:19:01+0000");")
        version_pattern = get_special_script_tag_pattern(
            SpecialScriptTag.VERSION
        )
        version_match = version_pattern.search(file_content)

        if not version_match:
            revision_pattern = re.compile(
                r'script_(?P<name>version)\s*\((?P<quote99>[\'"])?(?P<value>\$'
                r"Revision:\s*[0-9]{1,6}\s* \$)(?P=quote99)?\s*\)\s*;"
            )
            revision_match = revision_pattern.search(file_content)
            if not revision_match:
                yield LinterError(
                    "VT is missing script_version(); or is using a wrong "
                    "syntax.",
                    file=nasl_file,
                    plugin=self.name,
                )

        # script_tag(name:"last_modification",
        # value:"2019-03-21 12:19:01 +0000 (Thu, 21 Mar 2019)");
        last_modification_pattern = get_script_tag_pattern(
            ScriptTag.LAST_MODIFICATION
        )
        match_last_modified = last_modification_pattern.search(file_content)

        if not match_last_modified:
            old_pattern = re.compile(
                r'script_tag\(\s*name\s*:\s*(?P<quote>[\'"])(?P<name>'
                r"last_modification)(?P=quote)\s*,\s*value\s*:\s*(?P<quote2>"
                r'[\'"])?(?P<value>\$Date:\s*[A-Za-z0-9\:\-\+\,\s\(\)]{44}'
                r"\s*\$)(?P=quote2)?\s*\)\s*;"
            )
            match_last_modified = old_pattern.search(file_content)
            if not match_last_modified:
                yield LinterError(
                    "VT is missing script_tag("
                    'name:"last_modification" or is using a wrong '
                    "syntax.",
                    file=nasl_file,
                    plugin=self.name,
                )



# Copyright (C) 2021 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from pathlib import Path
from typing import Iterator

from troubadix.plugin import FileContentPlugin, LinterError, LinterResult


class AllScriptTagsPattern:
    instance = False

    def __init__(self) -> None:

        self.pattern = re.compile(
            pattern=(
                r'script_tag\(\s*name\s*:\s*(?P<quote>[\'"])'
                r"(?P<name>[a-zA-Z0-9\s\+\-\_]+)(?P=quote)\s*.+\s*\)\s*;"
            ),
            # flags=re.MULTILINE, # It seems that there is no multiline
            # script_tag here
        )
        self.instance = self


def get_all_tag_patterns() -> re.Pattern:
    """Get all script tags **without** matching the value!!!"""
    if AllScriptTagsPattern.instance:
        return AllScriptTagsPattern.instance.pattern
    return AllScriptTagsPattern().pattern


class CheckValidScriptTagNames(FileContentPlugin):
    """This step checks if the name of the following script tag:

    - script_tag(name:"", value:"");

    is one of the following allowed ones:

    - script_tag(name:"solution", value:"");
    - script_tag(name:"qod_type", value:"");
    - script_tag(name:"cvss_base", value:"");
    - script_tag(name:"cvss_base_vector", value:"");
    - script_tag(name:"summary", value:"");
    - script_tag(name:"last_modification", value:"");
    - script_tag(name:"insight", value:"");
    - script_tag(name:"affected", value:"");
    - script_tag(name:"creation_date", value:"");
    - script_tag(name:"vuldetect", value:"");
    - script_tag(name:"impact", value:"");
    - script_tag(name:"deprecated", value:"");
    - script_tag(name:"qod", value:"");
    - script_tag(name:"severity_vector", value:"");
    - script_tag(name:"severity_origin", value:"");
    - script_tag(name:"severity_date", value:"");
    - script_tag(name:"solution_method", value:"");
    # nb: Not fully implemented in GVM yet (further implementation "on hold").
    """

    name = "check_valid_script_tag_names"

    def check_content(
        self,
        nasl_file: Path,
        file_content: str,
    ) -> Iterator[LinterResult]:
        """
        Args:
            nasl_file: The VT that is going to be checked
        """
        if nasl_file.suffix == ".inc":
            return

        allowed_script_tag_names = [
            "solution",
            "solution_type",
            "qod_type",
            "cvss_base",
            "cvss_base_vector",
            "summary",
            "last_modification",
            "insight",
            "affected",
            "creation_date",
            "vuldetect",
            "impact",
            "deprecated",
            "qod",
            "severity_vector",
            "severity_origin",
            "severity_date",
            "solution_method",
        ]

        matches = get_all_tag_patterns().finditer(file_content)

        if matches:
            for match in matches:
                # print(match)
                if match.group("name") not in allowed_script_tag_names:
                    yield LinterError(
                        f"The script_tag name '{match.group('name')}' "
                        "is not allowed.",
                        file=nasl_file,
                        plugin=self.name,
                    )



#  Copyright (c) 2022 Greenbone Networks GmbH
#
#  SPDX-License-Identifier: GPL-3.0-or-later
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
import io
from contextlib import redirect_stdout
from pathlib import Path
from typing import Iterator

from codespell_lib import main as codespell_main
from troubadix.plugin import (
    FilePlugin,
    LinterError,
    LinterResult,
)

plugin_path = Path(__file__).parent.resolve()
codespell_config_path = (plugin_path.parent / "codespell").resolve()


class CheckSpelling(FilePlugin):
    name = "check_spelling"

    def run(self) -> Iterator[LinterResult]:
        """'codespell' is required to execute this step!
        This script opens a shell in a subprocess and executes 'codespell' to
        check the VT for spelling mistakes. An error will be thrown if
        'codespell' is not installed or corrections could be found via
        'codespell'.

        Args:
            nasl_file: The VT that is going to be checked
            _file_content: The content of the VT

        """

        # Overwrite with local repository files if exist
        if Path("codespell.additions").exists():
            codespell_additions = Path("codespell.additions").resolve()
        else:
            codespell_additions = f"{codespell_config_path}/codespell.additions"
        if Path("codespell.exclude").exists():
            codespell_exclude = Path("codespell.exclude").resolve()
        else:
            codespell_exclude = f"{codespell_config_path}/codespell.exclude"
        if Path("codespell.ignore").exists():
            codespell_ignore = Path("codespell.ignore").resolve()
        else:
            codespell_ignore = f"{codespell_config_path}/codespell.ignore"

        # Run codespell as internal process
        with redirect_stdout(io.StringIO()) as codespell:
            codespell_main(
                *(
                    "--hard-encoding-detection",
                    "--dictionary=-",
                    f"--dictionary={codespell_additions}",
                    f"--exclude-file={codespell_exclude}",
                    f"--ignore-words={codespell_ignore}",
                    "--disable-colors",
                    f"{str(self.context.nasl_file)}",
                )
            )

        codespell = codespell.getvalue()
        if "Traceback (most recent call last):" not in codespell:
            _codespell = codespell.splitlines()
            codespell = ""
            for line in _codespell:

                # From /Policy which is just a huge blob of text and too large
                # for codespell.exclude:
                if "policy_file_checksums_win.nasl" in line:
                    if re.search(r"nD\s+==>\s+and, 2nd", line) or re.search(
                        r"oD\s+==>\s+of", line
                    ):
                        continue

                # Same for a few other files:
                if "smtp_AV_42zip_DoS.nasl" in line and re.search(
                    r"BA\s+==>\s+BY, BE", line
                ):
                    continue

                if "bad_ssh_host_keys.inc" in line and re.search(
                    r"ba\s+==>\s+by, be", line
                ):
                    continue

                if "wmi_misc.inc" in line:
                    if re.search(r"BA\s+==>\s+BY, BE", line) or re.search(
                        r"OD\s+==>\s+OF", line
                    ):
                        continue

                if (
                    "ssl_funcs.inc" in line
                    or "gb_ssl_tls_cert_details.nasl" in line
                ):
                    if re.search(r"fpr\s+==>\s+for, far, fps", line):
                        continue

                # Name of a Huawei product
                if (
                    "gb_huawei" in line
                    or "telnetserver_detect_type_nd_version.nasl" in line
                ):
                    if re.search(
                        r"eSpace\s+==>\s+escape", line, flags=re.IGNORECASE
                    ):
                        continue

                # "ure" is a Debian package, again too many hits for
                # codespell.exclude.
                if re.search(
                    r"(deb_(dla_)?[0-9]+(_[" r"0-9]+)?|gb_ubuntu_.+)\.nasl",
                    line,
                ):
                    if re.search(r"ure\s+==>\s+sure", line):
                        continue

                # gsf/PCIDSS VTs are currently using some german text parts
                # nb: codespell seems to have some issues with german umlauts
                # in the codespell.exclude so a few of these were also excluded
                # here instead of directly via codespell.exclude.
                if (
                    "PCIDSS/" in line
                    or "GSHB/" in line
                    or "attic/PCIDSS_" in line
                    or "ITG_Kompendium/" in line
                ):
                    if re.search(
                        r"(sie|ist|oder|prozess|manuell|unter|funktion|"
                        r"alle|als|tage|lokale|uptodate|paket|titel|ba|"
                        r"ordner|modul|interaktive|programm|explizit|"
                        r"normale|applikation|attributen|lokal|signatur|"
                        r"modell|klick|generell)\s+==>\s+",
                        line,
                        flags=re.IGNORECASE,
                    ):
                        continue

                # False positives in the gsf/PCIDSS and GSHB/ VTs:
                # string('\nIn the file sent\nin milliseconds
                # There are too many hits to maintain them in codespell.exclude
                # so exclude them for now here.
                if (
                    "PCIDSS/" in line
                    or "GSHB/" in line
                    or "attic/PCIDSS_" in line
                ):
                    if re.search(r"n[iI]n\s+==>\s+inn", line):
                        continue

                # False positive in this VT in German example responses.
                if "gb_exchange_server_CVE-2021-26855_active.nasl" in line:
                    if re.search(r"ist\s+==>\s+is", line):
                        continue

                # Mostly a false positive in LSCs because of things like
                # "ALSA: hda" or a codec called "Conexant". There are too
                # many hits to maintain them in codespell.exclude so exclude
                # them for now here.
                if re.search(r"gb_(sles|(open)?suse)_.+\.nasl", line):
                    if re.search(
                        r"(hda|conexant)\s+==>\s+(had|connexant)",
                        line,
                        flags=re.IGNORECASE,
                    ):
                        continue

                # Jodie Chancel is a security researcher who is mentioned many
                # times in Mozilla advisories
                if re.search(
                    r"gb_mozilla_firefox_mfsa_\d{4}-\d{2,4}_lin\.nasl", line
                ) and re.search(r"Chancel\s+==>\s+Cancel", line):
                    continue

                # Look like correct as this is also in dictionary_rare.txt
                if "deb_dla_2896.nasl" in line and re.search(
                    r"dependant\s+==>\s+dependent", line
                ):
                    continue

                # Similar to the one above for SLES: "ALSA: hda". Also exclude
                # "tre", because it's a package name
                if re.search(r"mgasa-\d{4}-\d{4}.nasl", line) and re.search(
                    r"(hda|tre)\s+==>\s+(had|tree)", line, flags=re.IGNORECASE
                ):
                    continue

                # Similar to the corrections above, with some additional
                # exclusions like e.g. names
                if re.search(r"ELSA-\d{4}-\d{4,5}\.nasl", line):
                    if (
                        re.search(r"Stange\s+==>\s+Strange", line)
                        or re.search(r"chang\s+==>\s+change, charge", line)
                        or re.search(
                            r"IST\s+==>\s+IS, IT, ITS, IT'S, SIT, LIST", line
                        )
                        or re.search(r"hda\s+==>\s+had", line)
                        or re.search(r"Readded\s+==>\s+Read", line)
                        or re.search(r"ACI\s+==>\s+ACPI", line, re.IGNORECASE)
                        or re.search(r"UE\s+==>\s+USE, DUE", line)
                    ):
                        continue

                codespell += line + "\n"

        if "==>" in codespell:
            yield LinterError(
                codespell,
                file=self.context.nasl_file,
                plugin=self.name,
            )
        elif "Traceback (most recent call last):" in codespell:
            yield LinterError(
                codespell,
                file=self.context.nasl_file,
                plugin=self.name,
            )
